#!/bin/perl
#
##########################################################################
#
# Filename:     dns_update.pl
# Description:  Dynamic DNS-DHCP update scrypt
# Author:       Mike Stella
# Created:      November 10, 1998
# Last Updated: April 20, 2000
# Email:        kazin@sector13.org
# HomePage:     http://www.sector13.org/kazin
#
###########################################################################
#
# This code is Copyright (c) 1998,1999,2000 by Mike Stella
#
# NO WARRANTY is given for this program.  If it doesn't
# work on your system, sorry.  If it eats your hard drive, 
# again, sorry.  It works fine on mine.  Good luck!
#
###########################################################################
#
# This script reads a dhcpd.leases file and dynamically updates BIND with
# hostname-ip information.  
#
# It assumes that your DHCP server recieves hostnames from the
# clients, and that your clients offer their hostnames to the server.
# Some versions of Linux DHCP clients don't do that.  I use ISC's
# DHCPD, found at http://www.isc.org - though others may work just
# fine.
#
# As of version 0.1.3, you MUST use BIND 8.  This program now uses BIND's 
# dynamic update feature instead of rebuilding the databases and triggering
# a reload.  You also need the Net-DNS perl module, available from cpan.org.
#
###########################################################################
#
# 11/23/1998 - fixed the rarp stuff
# 02/23/1999 - added support for multiple subnets
# 02/24/1999 - improved lease file processing
# 04/15/1999 - switched to dynamic DNS update - BIND 8 dynamic capabilities
# 04/18/1999 - improved handling of client hostnames, to catch dumb user 
#              problems
#              improved error handling
#              Thanks to Wayne Roberts <wroberts1@home.com>
# 04/20/1999 - bugfix, error messages now show up correctly.
# 08/24/1999 - switched to using logger for logging to system log files
#              Thanks to Michael Matsumura <michael@limit.org>
# 08/27/1999 - Actually figured out how to update the reverse-lookup zones,
#              Thanks again to Michael Matsumura.
# 12/22/1999 - Now undefining hostname between runs to prevent creation of
#              duplicates if an address is requested without providing
#              hostname.
#              Thanks to Jasper Wallace <jasper@ivision.co.uk>
# 20/04/2000 - Added Bind8 Complianc, There are no Underscores if u Set
#	           bind8compl to Yes
#	           Thomas Bucher <Thomas@skypro.ch>	
# 20/04/2000 - Added No RAS feature, The Feature filters IP-Addresses 
#	           witch are taken from RAS interfaces. Just set NoRAS to Yes
#              Thomas Bucher <Thomas@Skypro.ch>
###########################################################################

use Net::DNS;
use strict;
$|=1;

###########################################################################
### Globals - you can change these as needed

# Domain name
my ($domain_name)        = "LAN";

# MAC address list - normally, you do not need this file
my ($mac_list)           = "/etc/dhcpd/mac_addresses.txt";

# DHCPD lease file
###my ($lease_file)         = "/var/state/dhcp/dhcpd.leases";
my ($lease_file)         = "/var/lib/dhcp/dhcpd.leases";

# number of seconds to check the lease file for updates
my ($update_freq)        = 90;

# Which nameserver do we update?
my ($nameserver)         = "192.168.1.2";

# Bind 8 Compliant - No _ in the the Workstation Name (Yes/No)
my ($bind8compl)	 = "Yes";

# Drop RAS IP-Addresses - Do not acctualize RAS-Ip addresses in the DNS (Yes/No)
my ($NoRAS)		 = "Yes";
###########################################################################
### Don't mess with anything below unless you REALLY need to modify the
### code.  And if you do, please let me know, I'm always interested in
### in improving this program.

# Make a pid file
`echo $$ > /var/run/dns_update.pid`;

my ($logstr);

# last modified time
my ($modtime) = 0;

use vars qw (%db);

my ($version) = "0.2.2.1";

###########################################################################
# Main Loop
while (1) {

  # check the file's last updated time, if it's been changed, update
  # the DNS and save the modified time.  This will ALWAYS run once - on
  # startup, since $modtime starts at zero.

  my (@stats) = stat ($lease_file);
  if ($stats[9] > $modtime) {

	# two date formats - one for logs, one to update the serial
	#$date = `date +\"%b %d %T\"`;
	#chomp $date;
	#$numdate = `date +\"%Y%j%H%M%S\"`;
	#chomp $numdate;

	# clear the old hash
	undef %db;

	`logger -t dns_update.pl dhcpd.leases changed - updating DNS`;
	$modtime = $stats[9];
	&read_lease_file;
	&update_dns;
 	#&print_db;
  } 

  # wait till next check time
  sleep $update_freq;

} # end main
###########################################################################


### update_dns - loops, setting each new IP into the DNS
sub update_dns {

  my ($ip, $hostname);
  while (($ip, $hostname) = each (%db)) {
	set_name ($hostname, $ip);
	set_rname ($hostname, $ip);
  }
}

### stick the new host into the DNS
sub set_name {
    my ($name,$ipaddress) = @_;
	$name .= '.' . $domain_name;
	$name = lc ($name);
	my ($update, $ans, $res, $query);
	
    # If there is already an A record for this name, remove it.
    $res = new Net::DNS::Resolver;
    $query = $res->search($name);
	my $rr;
    if ($query) {
	  foreach $rr ($query->answer) {
		next unless $rr->type eq "A";
	  }
	  $update = new Net::DNS::Update($domain_name);
	  
	  # Prerequisite (assumed) is that an A record must already exist.
	  $update->push("update", new Net::DNS::RR(Name    => $name,
											   Type    => "A",
											   Ttl     => 0,
											   Class   => "ANY",
											   Rdata   => ""));
	  
	  $res = new Net::DNS::Resolver;
	  $res->nameservers($nameserver);
	  my $ans = $res->send($update);
	  
	  if (defined $ans) {
		if ($ans->header->rcode ne "NOERROR") {
		  $logstr = "Old entry removal failed: " . $ans->header->rcode;
		  `logger -t dns_update.pl $logstr`;
		}
	  }
	  else {
		$logstr = "Old entry removal failed: " . $res->errorstring;
	  }
	  $logstr = "Replacing $name in zone $domain_name - ";
    }
	else {
	  $logstr = "Adding $name to zone $domain_name - ";
	}
	
    $update = new Net::DNS::Update($domain_name);
    # NXRRSET - Prerequisite is that no A records exist for the name.
    $update->push("pre", new Net::DNS::RR(Name  => $name,
                                          Class => "NONE",
                                          Type  => "A"));

    # Add one A records for the name.
    $update->push("update", new Net::DNS::RR(Name    => $name,
                                             Ttl     => 10,
                                             Type    => "A",
                                             Address => $ipaddress));

    $res = new Net::DNS::Resolver;
    $res->nameservers($nameserver);
    $ans = $res->send($update);

    if (defined $ans) {
	  if ($ans->header->rcode eq "NOERROR") {
		$logstr .= "Success.";
	  }
	  else {
		$logstr .= "Failed: " . $ans->header->rcode;
	  }
    }
	else {
	  $logstr .= "Failed: " . $res->errorstring;
    }
    `logger -t dns_update.pl $logstr`;
}

### stick the new host into the DNS (PTR)
sub set_rname {
  my ($name,$ipaddress) = @_;
  $name .= '.' . $domain_name;
  $name = lc ($name);
	
  my $ipend = $ipaddress;
  $ipend =~ s/(\d+)\.(\d+)\.(\d+)\.//g;
  
  my $rdomain = $ipaddress;
  $rdomain =~ s/^(\d+)\.(\d+)\.(\d+)\.(\d+)$/$3.$2.$1.in-addr.arpa/;
  my $arpaaddr = $ipend . "." . $rdomain;
  
  my ($update, $ans, $res, $query);
  
  # If there is already an A record for this name, remove it.
  $res = new Net::DNS::Resolver;
  $query = $res->search($ipaddress);
  my $rr;
  if ($query) {
	#	  print $logstr . "Replacing $ipaddress = ";
	#	  `logger -t dns_update.pl "Replacing $ipaddress = "`;
	foreach $rr ($query->answer) {
	  next unless $rr->type eq "PTR";
	}
	$update = new Net::DNS::Update($rdomain);
	
	# Prerequisite (assumed) is that an A record must already exist.
	$update->push("update", new Net::DNS::RR(Name    => $arpaaddr,
											 Type    => "PTR",
											 Ttl     => 0,
											 Class   => "ANY",
											 Rdata   => ""));
	
	$res = new Net::DNS::Resolver;
	$res->nameservers($nameserver); 
	my $ans = $res->send($update);
	
	if (defined $ans) {
	  if ($ans->header->rcode ne "NOERROR") {
		$logstr = "Old entry removal failed: " . $ans->header->rcode;
		`logger -t dns_update.pl $logstr`;
	  }
	}
	else {
	  $logstr = "Old entry removal failed: ". $res->errorstring;
	}
	$logstr = "Replacing $ipaddress in zone $rdomain - ";
  }
  else {
	$logstr = "Adding $ipaddress to zone $rdomain - ";
  }
  
  $update = new Net::DNS::Update($rdomain);
  # NXRRSET - Prerequisite is that no A records exist for the name.
  $update->push("pre", new Net::DNS::RR(Name  => $arpaaddr,
										Class => "NONE",
										Type  => "PTR"));
  
  # Add one A records for the name.
# $update->push("update", new Net::DNS::RR(Name    => $arpaaddr,
#										   Ttl     => 10,
#										   Type    => "PTR",
#										   Address => $name));
 $update->push("update", new Net::DNS::RR("$arpaaddr 10 PTR $name"));
  
  $res = new Net::DNS::Resolver;
  $res->nameservers($nameserver);
  $ans = $res->send($update);
  
  if (defined $ans) {
	if ($ans->header->rcode eq "NOERROR") {
	  $logstr .= "Success.";
	}
	else {
	  $logstr .= "Failed: " . $ans->header->rcode;
	}
  }
  else {
	$logstr .= "Failed: " . $res->errorstring;
  }
  `logger -t dns_update.pl $logstr`;

}

### read_lease_file
### reads the lease file & makes a hash of what's in there.
sub read_lease_file {

  unless (open(LEASEFILE,$lease_file)) {
	`logger -t dns_update.pl error opening dhcpd lease file`;
	return;
  }

  my (@tmp, $ip, $hostname, $mac);
  ## Loop here, reading from LEASEFILE
  while (<LEASEFILE>) {
	if (/^\s*lease/i) {
	  # find ip address
	  @tmp = split /\s+/;
	  $ip = $tmp[1];
	  
	  # do the rest of the block - we're only interested in hostname
	  # and mac address
	  while ($_ !~ /^}/) {
	    $_ = <LEASEFILE>;
	    if ($_ =~ /^\s*hardware/i) {
		# find mac address
		chomp $_;
		chop $_;
		@tmp = split /\s+/;
		$mac = $tmp[3];
		}
		elsif ($_ =~ /^\s*client/i) {
		  chomp $_;
		  chop $_;
		  # find hostname
		  $_ =~ /\"(.*)\"/;
		  $hostname = $1;
		  # change spaces to underscore, remove dots
		  $hostname =~ s/\s/_/g;
		  $hostname =~ s/\.//g;
		  if ($bind8compl == "Yes") {
		    $hostname =~ s/\_/-/g;
		  }
		}
	  }

	  # grab hostname from mac list file if it's not provided by
	  # the DHCP client.
	  if (($hostname eq "") and ($mac ne "")) {
		$hostname = &mac_to_hostname($mac);
	  }

	  # Store hostname/ip in hash
	  unless (($NoRAS eq "Yes") and ((length $mac) >=30)) { 
             unless ($hostname eq "") {
		$db{$ip} = $hostname;
		# print "$ip - $hostname - $mac\n"; 
	        # DEBUG of Database with MAC address
	     }
          }
	  $hostname = undef;
	}
  }

  close LEASEFILE;
}

### left around for testing
sub print_db {
  my ($key,$value);

  while (($key,$value) = each (%db)) {
	print "$key - $value\n";
  }
}

### mac_to_hostname
### give this a MAC address, and it gives you the hostname.
sub mac_to_hostname {
  my ($mac) = @_;

  # Make sure the file is there...
  unless( -f $mac_list) {
	`logger -t dns_update.pl error opening mac_list`;
	return;
  }
  
  # grab the hostname, and return it.
  my ($mac_addr,$hostname) = split /\s+/, `grep -i $mac $mac_list`;
  return $hostname;
}

